import re
import pandas as pd
from scipy import stats

from parse_gtf import get_gene_df, get_gene_merge_exon_dic, gene_df2dic
from plot import boxplot, scatterplot
gtf = '../../../data/Homo_sapiens.GRCh38.93.gtf'
gene_df = get_gene_df( gtf )
filter_gene_df = gene_df[ gene_df['genetype'].isin(['protein_coding', 'lincRNA', 'lncRNA']) ]
print( filter_gene_df[filter_gene_df['gene_name']=='MIR1302-2HG'])
print( filter_gene_df.shape )
print( set(filter_gene_df['genetype']) )
gene_range_dic = gene_df2dic(filter_gene_df)

chr_list = list( set( filter_gene_df['chrom'] ) )

sort_chr_list = [chrom for chrom in chr_list if re.match( r'chr\d+', chrom)]
sort_chr_list.sort(key=lambda arr: (arr[:3], int(arr[3:])))
if 'chrM' in chr_list:  sort_chr_list.append('chrM')
if 'chrMT' in chr_list:  sort_chr_list.append('chrMT')
if 'chrX' in chr_list:  sort_chr_list.append('chrX')
if 'chrY' in chr_list:  sort_chr_list.append('chrY')

output_root = '../../../nasap_GSM3618143/'
attr_df = pd.read_csv(output_root + 'csv/feature_attrs.csv', index_col=0)
rpkm_series = attr_df['rpkm']
gene_rpkm_dic = rpkm_series.to_dict()
print( dict(list(gene_rpkm_dic.items())[:3]) )

rpkm_list = []
for chrom in sort_chr_list:
  gene_list = filter_gene_df[ filter_gene_df['chrom'] == chrom ]['gene_name']

  cur_rpkm_list = []
  for gene in gene_list:
    try:
      cur_rpkm_list.append(gene_rpkm_dic[gene])
    except:
      continue
  rpkm_list.append( cur_rpkm_list )


boxplot(rpkm_list, 'chromsome genes RPKM', 'chromsome', 'RPKM', sort_chr_list, output_root + 'imgs/chr_rpkm.png' )


exon_intron_df = pd.read_csv(output_root + 'csv/exon_intron_ratio.csv', index_col=0)
filter_exon_intron_df = exon_intron_df[(exon_intron_df['exon'] > 0) & (exon_intron_df['intron'] >0) ]

slope, intercept, r_value, p_value, std_err = stats.linregress(filter_exon_intron_df['exon'],filter_exon_intron_df['intron'])
text= 'y = {:4.2e} x + {:4.2e}; \nR^2= {:2.2f}'.format(slope, intercept, r_value*r_value)
scatterplot( filter_exon_intron_df['exon'], filter_exon_intron_df['intron'], 'Exon/Intron ratio', 'Exon', 'Intron', text, output_root+'imgs/exon_intron_ratio.png' )



import matplotlib.pyplot as plt
import matplotlib.colors as colors
color_list = list(colors._colors_full_map.values())

def boxplot( boxplot_data, title, x_label, y_label, xtick_labels, output_file ):
  # matplotlib比较old， 输入是 二维list
  fig = plt.figure(1, figsize=(12, 7))
  ax = fig.add_subplot(111)
  bp = ax.boxplot(boxplot_data, patch_artist=True)

  for box in bp['boxes']:
    box.set(facecolor='#087E8B', alpha=0.6, linewidth=2)

  for whisker in bp['whiskers']:
    whisker.set(linewidth=2)

  for median in bp['medians']:
    median.set(color='black', linewidth=3)

  ax.set_title( title )
  ax.set_xlabel( x_label)
  ax.set_ylabel( y_label )
  ax.set_xticklabels( xtick_labels )
  fig.savefig(output_file)
  plt.close(fig)

def scatterplot( x_list, y_list, title, x_label, y_label, text, output_file ):
  fig = plt.figure(figsize=(12, 7))
  ax = plt.gca()
  plt.scatter(x=x_list, y=y_list)
  plt.text(0.5, 0.75, text, transform = ax.transAxes)
  plt.title( title )
  plt.xlabel(x_label)
  plt.ylabel(y_label)
  fig.savefig(output_file)
  plt.close(fig)

def lollipopplot(chrs, lollipop_dic, output_file):
  fig, axes = plt.subplots(len(chrs)*2, 1, sharex=True, sharey=False, figsize=(30, 2*len(chrs) ))
  plt.xticks([])
  plt.subplots_adjust(wspace =0, hspace =0)
  i = 0
  for chr in chrs:
    color = color_list[i]
    try:
      axes[i].stem( lollipop_dic[chr]['forward']['site'], lollipop_dic[chr]['forward']['count'], linefmt=color, markerfmt='.', basefmt=color  )
    except:
      pass
    axes[i].set_ylabel( chr+':+', color=color )
    axes[i].yaxis.tick_right()
    axes[i].yaxis.get_major_ticks()[1].set_visible(False)
    axes[i].yaxis.get_major_ticks()[-2].set_visible(False)
    axes[i].spines['top'].set_visible(False)
    axes[i].spines['left'].set_visible(False)
    axes[i].spines['right'].set_color( color )
    i+=1
    try:
      axes[i].stem( lollipop_dic[chr]['reverse']['site'], [-1*x for x in lollipop_dic[chr]['reverse']['count']], linefmt=color, markerfmt='.', basefmt=color )
    except:
      pass
    axes[i].set_ylabel( chr+':-', color=color)
    axes[i].yaxis.tick_right()
    axes[i].yaxis.get_major_ticks()[1].set_visible(False)
    axes[i].yaxis.get_major_ticks()[-2].set_visible(False)
    axes[i].spines['bottom'].set_visible(False)
    axes[i].spines['left'].set_visible(False)
    axes[i].spines['right'].set_color(color)
    i+=1


  plt.title('Genome wide pausing sites distribution', y=1.1, fontsize=18)
  fig.savefig(output_file)
  plt.close(fig)


import os, sys, fire, psutil, time

from loguru import logger

# from .libs.py_ext import
# 思路:
# 每个命令都有参数，参数是值或文件(文件夹), 每个参数都要被检验
# 参数可以是required 或者 非required
# 如果是文件则检查文件是否存在
# 文件夹如果不存在，并创建


self_scripts = os.path.abspath( os.path.join(os.path.dirname(__file__)) ) + '/scripts/'


def _check_soft(name_list, isPython=False):
  if isPython:
    for name in name_list:
      try:
        exec('import '+ name )
      except:
        logger.error( 'import %s error.'%name)
        if name == 'community':
          name = 'python-louvain'
        logger.error( 'you need to install %s '%name)
        os.sys.exit(1)
  else:
    for name in name_list:
      soft_dep = os.system(name + ' --version')
      if soft_dep != 0:
        logger.error( 'import %s error.'%name)
        logger.error( 'you need to install %s first'%name)
        os.sys.exit(1)

def _check_para(name, value, type, isFile=False, required=False, output_root='./test_output/'):
  # 先检查参数的类型，
  # 如果是文件就检查文件是否存在
  # 如果是文件夹就判断文件夹是否存在，不存在就创建
  if not output_root.endswith('/'): output_root = output_root + '/'
  if required:
    if value == None:
      logger.error( name + ' is a required parameter')
      with open( output_root + 'error.log', 'w') as f:
        f.write( name + ' is a required parameter' )
      os.sys.exit(0)
  if not isinstance(value, type):
    logger.error('Error'+ name + 'should be a', type, 'type.')
    with open( output_root + 'error.log', 'w') as f:
      f.write( 'Error', name, 'should be a', type, 'type.')
    os.sys.exit(0)
  if isFile:
    if not os.path.exists(value):
      logger.error('Error' + value + 'not exist!')
      with open( output_root + 'error.log', 'w') as f:
        f.write( 'Error'+ value + 'not exist!')
      os.sys.exit(0)
  return True

def _get_cores(value):
  total_cores = psutil.cpu_count()
  cur_cores = 1
  if isinstance(value, int):
    if value > total_cores:
      print("warning: your cpu cores number is", total_cores, "we set the parameter cores to", total_cores)
      cur_cores = total_cores
    else:
      cur_cores = value
    return cur_cores
  if isinstance(value, str):
    if str.lower() == 'max':
      cur_cores = total_cores
      return cur_cores
    if str.lower == 'max/2':
      cur_cores = total_cores/2
      return cur_cores
  print("warning: cores parameter error use default cores=1")
  return 1

def _cur_time():
  return time.asctime( time.localtime(time.time()) )

def _build_project_path(root_dir):
  if not root_dir.endswith('/'): root_dir = root_dir + '/'
  json_dir=root_dir + 'json/'
  fq_dir=root_dir + 'fastq/'
  txt_dir=root_dir + 'txt/'
  img_dir=root_dir + 'imgs/'
  sam_dir=root_dir + 'sam/'
  bed_dir=root_dir + 'bed/'
  bw_dir=root_dir + 'bw/'
  csv_dir=root_dir + 'csv/'
  html_dir=root_dir + 'html/'
  for d in [root_dir, json_dir, fq_dir, txt_dir, img_dir, sam_dir, bed_dir, bw_dir, csv_dir, html_dir]:
    if not os.path.exists(d):
      os.makedirs(d)

def _check_output(root_dir):
  output_subdir = ['json/','fastq/','txt/', 'imgs/','sam/','bed/','bw/', 'html/']

  for sub_dir in output_subdir:
    if not root_dir.endswith('/'): root_dir = root_dir + '/'
    sub_dir = root_dir + sub_dir
    if not os.listdir(sub_dir):
      os.removedirs(sub_dir)


def server(output_root='./test_output/', forward_bw=None, reverse_bw=None, gtf=None,  cores=1, tf_source=None, tf_filter_nodes=None, enhancer_source=None, enhancer_filter_nodes=None):
  logger.info('server--check installed software')
  _check_soft( ['pandas', 'pyBigWig', 'numpy', 'scipy', 'networkx','community', 'hvplot'], isPython=True)

  logger.info('server--construct output dir')
  if not output_root.endswith('/'): output_root = output_root + '/'
  _build_project_path(output_root)

  logger.info('server--check parameter and input files')

  feature_assign_cmd_list = ['python', self_scripts + 'feature_attrs.py']
  pausing_sites_cmd_list = ['python', self_scripts + 'pausing_sites_low_memory.py']
  network_cmd_list = ['python', self_scripts + 'network_analysis.py']
  render_cmd_list = ['python', self_scripts + 'render_template.py']
  render_cmd_list.extend(['--type', 'server'])

  feature_assign_cmd_list.extend(['--output_root', output_root])
  pausing_sites_cmd_list.extend(['--output_root', output_root])
  network_cmd_list.extend(['--output_root', output_root])
  render_cmd_list.extend(['--output_root', output_root])

  # 检查必须参数
  # 检查文件
  _check_para('--forward_bw', forward_bw, str, isFile=True, required=True, output_root=output_root)
  feature_assign_cmd_list.extend(['--forward_bw', forward_bw] )
  pausing_sites_cmd_list.extend(['--forward_bw', forward_bw] )
  _check_para('--reverse_bw', reverse_bw, str, isFile=True, required=True, output_root=output_root)
  feature_assign_cmd_list.extend(['--reverse_bw', reverse_bw] )
  pausing_sites_cmd_list.extend(['--reverse_bw', reverse_bw] )
  _check_para('--gtf', gtf, str, isFile=True, required=True, output_root=output_root)
  feature_assign_cmd_list.extend(['--gtf', gtf] )

  # 可选参数
  cur_cores = _get_cores(cores)
  pausing_sites_cmd_list.extend(['--cores', str(cur_cores)])

  if not (tf_source or enhancer_source):
    logger.error( 'Error, you have to provide at least one parameter for tf_source or enhancer_source')
    os.sys.exit(1)

  if tf_source:
    _check_para('--tf_source', tf_source, str, isFile=True, required=True, output_root=output_root)
    network_cmd_list.extend( ['--tf_source', tf_source])
  if tf_filter_nodes:
    _check_para('--tf_filter_nodes', tf_filter_nodes, str, isFile=True, required=True, output_root=output_root)
    network_cmd_list.extend( ['--tf_filter_nodes', tf_filter_nodes])
  if enhancer_source:
    _check_para('--enhancer_source', enhancer_source, str, isFile=True, required=True, output_root=output_root)
    network_cmd_list.extend( ['--enhancer_source', enhancer_source])
  if enhancer_filter_nodes:
    _check_para('--enhancer_filter_nodes', enhancer_filter_nodes, str, isFile=True, required=True, output_root=output_root)
    network_cmd_list.extend( ['--enhancer_filter_nodes', enhancer_filter_nodes])

  all_steps = [feature_assign_cmd_list, pausing_sites_cmd_list, network_cmd_list, render_cmd_list]
  steps_name = ['feature_assign', 'pausing_sites', 'network_analysis', 'render_output']
  for cmd_list, step in zip(all_steps, steps_name):
    try:
      logger.info(step + ' start')
      os.system( ' '.join(cmd_list) )
      logger.info(step + ' finished')
    except Exception as e:
      with open( output_root + 'error.log', 'w') as f:
        logger.error( e )
        f.write(e)
      os.sys.exit(1)

def assessment(output_root='./test_output/', read1=None,  cores=1, read2=None, adapter1=None, adapter2=None, umi=None, bowtie_index=None, gtf=None):
  logger.info('assessment--check installed software')
  _check_soft( ['fastp', 'bioawk', 'python', 'bowtie2', 'samtools', 'bedtools', 'deeptools'], isPython=False)

  logger.info('assessment--construct output dir')
  if not output_root.endswith('/'): output_root = output_root + '/'
  _build_project_path(output_root)

  # 这里不检查 文件是否存在，因为在_preprocess函数中有检查
  _preprocess(output_root=output_root, read1=read1, read2=read2, adapter1=adapter1, adapter2=adapter2, umi=umi )
  if read2:
    _alignment(output_root=output_root, read1=output_root + 'fastq/clean_read1.fq.gz', read2=output_root + 'fastq/clean_read2.fq.gz', bowtie_index=bowtie_index, gtf=gtf, cores=cores )
  else:
    _alignment(output_root=output_root, read1=output_root + 'fastq/clean_read1.fq.gz', bowtie_index=bowtie_index, gtf=gtf, cores=cores )

  _genome_tracks(output_root=output_root, bam=output_root + 'sam/uniquemapped_sort.bam')
  _render_template(output_root=output_root, type='assessment', is_server='False')

def all(output_root='./test_output/', read1=None, bowtie_index=None,  gtf=None,  cores=1, read2=None, adapter1=None, adapter2=None, umi=None, tf_source=None, tf_filter_nodes=None, enhancer_source=None, enhancer_filter_nodes=None):
  logger.info('all--check installed software')
  _check_soft( ['fastp', 'bioawk', 'python', 'bowtie2', 'samtools', 'bedtools', 'deeptools'], isPython=False)
  _check_soft( ['pandas', 'pyBigWig', 'numpy', 'scipy', 'networkx','community', 'hvplot'], isPython=True)
  if not output_root.endswith('/'): output_root = output_root + '/'
  _build_project_path(output_root)
  assessment(output_root=output_root, read1=read1,  cores=cores, read2=read2, adapter1=adapter1, adapter2=adapter2, umi=umi, bowtie_index=bowtie_index, gtf=gtf)
  server(output_root=output_root, forward_bw=output_root + 'bw/forward.bw', reverse_bw=output_root + 'bw/reverse.bw', gtf=gtf, cores=cores, tf_source=tf_source, tf_filter_nodes=tf_filter_nodes, enhancer_source=enhancer_source, enhancer_filter_nodes=enhancer_filter_nodes )

  # 把模块组合
  # preprocess_cmd_list = ['bash', self_scripts + 'preprocess.bash']
  # extract_preprocess_cmd_list = ['python', self_scripts + 'extract_preprocess.py']
  # map1_cmd_list = ['bash', self_scripts + 'map1.bash']
  # map2_cmd_list = ['python', self_scripts + 'map_split.py', '--sam_file=original.sam']
  # map3_cmd_list = ['bash', self_scripts + 'map2.bash']
  # genome_track_cmd_list = ['bamCoverage', '--binSize', '1']
  # feature_assign_cmd_list = ['python', self_scripts + 'feature_attrs.py']
  # pausing_sites_cmd_list = ['python', self_scripts + 'pausing_sites.py']

  # # 必须参数 参数什么都不加
  # _check_para('--read1', read1, str, isFile=True, required=True, output_root=output_root)
  # preprocess_cmd_list.extend(['--read1', read1])

  # _check_para('--bowtie_index', bowtie_index, str, required=True, output_root=output_root)
  # map1_cmd_list.extend(['--bowtie_index', bowtie_index])
  # map3_cmd_list.extend(['--bowtie_index', bowtie_index])

  # _check_para('--gtf', gtf, str, isFile=True, required=True, output_root=output_root)
  # map1_cmd_list.extend(['--gtf', gtf])
  # map3_cmd_list.extend(['--gtf', gtf])
  # feature_assign_cmd_list.extend(['--gtf', gtf])

  # if output_root:
  #   if not output_root.endswith('/'): output_root = output_root +'/'
  #   preprocess_cmd_list.extend(['--output_root', output_root])
  #   extract_preprocess_cmd_list.append(output_root)
  #   map1_cmd_list.extend(['--output_root', output_root])
  #   map1_cmd_list.extend(['--read1', output_root + 'fastq/clean_read1.fq.gz'])
  #   map2_cmd_list.append('--sam_dir='+output_root+'sam/')
  #   map3_cmd_list.extend(['--output_root', output_root])
  #   map3_cmd_list.extend(['--read1', output_root + 'fastq/clean_read1.fq.gz'])
  #   genome_track_cmd_list.extend(['--bam', output_root + 'sam/uniquemapped_sort.bam'])
  #   feature_assign_cmd_list.extend(['--forward_bw', output_root + 'bw/forward.bw'] )
  #   feature_assign_cmd_list.extend(['--reverse_bw', output_root + 'bw/reverse.bw'] )
  #   feature_assign_cmd_list.extend(['--output_root', output_root])
  #   pausing_sites_cmd_list.extend(['--forward_bw', output_root + 'bw/forward.bw'] )
  #   pausing_sites_cmd_list.extend(['--reverse_bw', output_root + 'bw/reverse.bw'] )
  #   pausing_sites_cmd_list.extend(['--output_root', output_root])
  # if cores:
  #   cur_cores = str( _get_cores(cores) )
  #   preprocess_cmd_list.extend(['--cores', cur_cores])
  #   map1_cmd_list.extend(['--cores', cur_cores])
  #   map3_cmd_list.extend(['--cores', cur_cores])
  #   genome_track_cmd_list.extend(['-p', cur_cores])
  #   pausing_sites_cmd_list.extend(['--cores', cur_cores])

  # # 可选参数
  # if read2:
  #   _check_para('--read2', read2, str, isFile=True, required=True, output_root=output_root)
  #   preprocess_cmd_list.extend(['--read2', read2])
  #   clean_read2=output_root + 'fastq/clean_read2.fq.gz'
  #   map1_cmd_list.extend(['--read2', clean_read2])
  #   map3_cmd_list.extend(['--read2', clean_read2])
  # if adapter1:
  #   _check_para('--adapter1', adapter1, str, required=True, output_root=output_root)
  #   preprocess_cmd_list.extend(['--adapter1', adapter1])
  # if adapter2:
  #   _check_para('--adapter2', adapter2, str, required=True, output_root=output_root)
  #   preprocess_cmd_list.extend(['--adapter2', adapter2])

  # foward_genome_track_cmd_list, reverse_genome_track_cmd_list = genome_track_cmd_list.copy(), genome_track_cmd_list.copy()
  # foward_genome_track_cmd_list.extend(['--filterRNAstrand', 'forward'])
  # reverse_genome_track_cmd_list.extend(['--filterRNAstrand', 'reverse'])
  # foward_genome_track_cmd_list.extend(['-o', output_root + 'bw/forward.bw'])
  # reverse_genome_track_cmd_list.extend(['-o', output_root + 'bw/reverse.bw'])

  # all_steps = [preprocess_cmd_list, extract_preprocess_cmd_list, map1_cmd_list, map2_cmd_list, map3_cmd_list,
  # foward_genome_track_cmd_list, reverse_genome_track_cmd_list, pausing_sites_cmd_list]
  # _build_project_path(output_root)
  # for cmd_list in all_steps:
  #   os.system( ' '.join(cmd_list) )
    # print( ' '.join(cmd_list) )

  # _check_output(output_root)



def _preprocess(output_root=os.getcwd()+'/tmp_output', read1=None,  cores=1, read2=None, adapter1=None, adapter2=None, umi=None):
  logger.info('preprocess--check installed software')
  _check_soft( ['fastp', 'bioawk', 'python'], isPython=False)
  if read2:
    _check_soft( ['flash'], isPython=False)

  logger.info('preprocess--check parameter and input files')
  if not output_root.endswith('/'):
    output_root = output_root + '/'
  logger.info('preprocess--construct output dir')
  _build_project_path(output_root)
  log_file = open(output_root + 'tmp.log', 'w')
  cmd_list = ['bash', self_scripts + 'preprocess.bash']
  cmd_list.extend(['--output_root', output_root])
  # 必须参数 参数什么都不加
  _check_para('--read1', read1, str, isFile=True, required=True, output_root=output_root)
  cmd_list.extend(['--read1', read1])
  log_file.write( 'read1--'+os.path.basename(read1)+'\n' )
  log_file.write('read1_size--' + str(os.stat(read1).st_size / (1024 * 1024)) + 'Mb'+'\n')

  # 可选参数
  cur_cores = _get_cores(cores)
  cmd_list.extend(['--cores', str(cur_cores)])

  if read2:
    _check_para('--read2', read2, str, isFile=True, required=True, output_root=output_root)
    cmd_list.extend(['--read2', read2])
    log_file.write( 'read2--'+os.path.basename(read2)+'\n' )
    log_file.write('read2_size--' + str(os.stat(read2).st_size / (1024 * 1024)) + 'Mb'+'\n')

  if adapter1:
    _check_para('--adapter1', adapter1, str, required=True, output_root=output_root)
    cmd_list.extend(['--adapter1', adapter1])
    log_file.write( 'adapter1--'+ adapter1 +'\n' )


  if adapter2:
    _check_para('--adapter2', adapter2, str, required=True, output_root=output_root)
    cmd_list.extend(['--adapter2', adapter2])
    log_file.write( 'adapter2--'+ adapter2 +'\n' )

  # print( ' '.join(cmd_list) )
  log_file.write(_cur_time()+ 'preprocess start'+'\n')
  logger.info('preprocess--running script')
  os.system( ' '.join(cmd_list) )
  # extract preprocess
  os.system( ' '.join(['python', self_scripts + 'extract_preprocess.py', output_root]) )

  log_file.close()
  logger.success(_cur_time() + 'preprocess--Finished. Find the results in ' + output_root)

def _alignment(read1=None, bowtie_index=None, gtf=None, output_root=None, cores=1, read2=None ):
  logger.info('alignment--check installed software')
  _check_soft( ['bowtie2', 'samtools', 'bedtools'], isPython=False)

  logger.info('alignment--check parameter and input files')
  if not output_root.endswith('/'): output_root = output_root + '/'
  logger.info('alignment--construct output dir')
  _build_project_path(output_root)
  log_file = open(output_root + 'tmp.log', 'a+' )
  cmd_list = ['bash', self_scripts + 'map1.bash']
  cmd_list.extend(['--output_root', output_root])

  _check_para('--read1', read1, str, isFile=True, required=True, output_root=output_root)
  cmd_list.extend(['--read1', read1])
  log_file.write('clean_read1--'+os.path.basename(read1) + '\n' )
  log_file.write('clean_read1_size--' + str(os.stat(read1).st_size / (1024 * 1024)) + 'Mb\n')

  _check_para('--bowtie_index', bowtie_index, str, required=True, output_root=output_root)
  cmd_list.extend(['--bowtie_index', bowtie_index])

  _check_para('--gtf', gtf, str, isFile=True, required=True, output_root=output_root)
  cmd_list.extend(['--gtf', gtf])


  cur_cores = _get_cores(cores)
  cmd_list.extend(['--cores', str(cur_cores)])
  if read2:
    _check_para('--read2', read2, str, isFile=True, required=True, output_root=output_root)
    cmd_list.extend(['--read2', read2])
    log_file.write('clean_read2--'+os.path.basename(read2) + '\n' )
    log_file.write('clean_read2_size--' + str(os.stat(read2).st_size / (1024 * 1024)) + 'Mb\n')

  log_file.write(_cur_time()+ 'alignment start\n')
  os.system( ' '.join(cmd_list) )
  os.system( ' '.join(['python', self_scripts + 'map_split.py',
  '--sam_dir='+output_root+'sam/', '--sam_file=original.sam',
  '--output_root='+output_root
  ]) )
  cmd_list[1] = self_scripts + 'map2.bash'
  os.system( ' '.join(cmd_list) )
  log_file.close()
  logger.success(_cur_time()+ 'alignment--Finished. Find the results in ' + output_root)

def _genome_tracks(bam=None, output_root=None, cores=1):
  logger.info('genome_tracks--check installed software')
  _check_soft( ['deeptools'], isPython=False)

  logger.info('genome_tracks--check parameter and input files')
  if not output_root.endswith('/'): output_root = output_root + '/'
  logger.info('genome_tracks--construct output dir')
  _build_project_path(output_root)
  log_file = open(output_root + 'tmp.log', 'a+' )
  cmd_list = ['bamCoverage', '--binSize', '1']
  _check_para('--bam', bam, str, isFile=True, required=True, output_root=output_root)
  cmd_list.extend(['--bam', bam])
  log_file.write('bam--'+os.path.basename(bam) + '\n' )
  log_file.write('bam_size--' + str(os.stat(bam).st_size / (1024 * 1024)) + 'Mb\n')
  # 这里运行时间太长 cores直接写死 40，40是线程
  # if cores:
  #   cur_cores = _get_cores(cores)
  #   cmd_list.extend(['-p', str(cur_cores)])
  cmd_list.extend(['-p', str(40)])
  cmd_list_forward, cmd_list_reverse = cmd_list.copy(), cmd_list.copy()
  cmd_list_forward.extend(['--filterRNAstrand', 'forward'])
  cmd_list_reverse.extend(['--filterRNAstrand', 'reverse'])
  cmd_list_forward.extend(['-o', output_root + 'bw/forward.bw'])
  cmd_list_reverse.extend(['-o', output_root + 'bw/reverse.bw'])
  log_file.write(_cur_time()+ 'generate forward genome track start\n')
  os.system( ' '.join(cmd_list_forward) )
  log_file.write(_cur_time()+ 'generate reverse genome track start\n')
  os.system( ' '.join(cmd_list_reverse) )
  log_file.close()
  logger.success(_cur_time()+'genome_tracks--Finished. Find the results in ' + output_root)

def feature_assign(forward_bw=None, reverse_bw=None, gtf=None, output_root=None):
  logger.info('feature_assign--check installed software')
  _check_soft( ['pandas', 'pyBigWig', 'scipy'], isPython=True)

  logger.info('feature_assign--check parameter and input files')
  if not output_root.endswith('/'): output_root = output_root + '/'
  logger.info('feature_assign--construct output dir')
  _build_project_path(output_root)
  log_file = open(output_root + 'tmp.log', 'w' )

  cmd_list = ['python', self_scripts + 'feature_attrs.py']
  _check_para('--forward_bw', forward_bw, str, isFile=True, required=True, output_root=output_root)
  cmd_list.extend(['--forward_bw', forward_bw])
  log_file.write('forward_bw--'+os.path.basename(forward_bw) + '\n' )
  log_file.write('forward_bw_size--' + str(os.stat(forward_bw).st_size / (1024 * 1024)) + 'Mb\n')

  _check_para('--reverse_bw', reverse_bw, str, isFile=True, required=True, output_root=output_root)
  cmd_list.extend(['--reverse_bw', reverse_bw])
  log_file.write('reverse_bw--'+os.path.basename(reverse_bw) + '\n' )
  log_file.write('reverse_bw_size--' + str(os.stat(reverse_bw).st_size / (1024 * 1024)) + 'Mb\n')

  _check_para('--gtf', gtf, str, isFile=True, required=True, output_root=output_root)
  cmd_list.extend(['--gtf', gtf])

  cmd_list.extend(['--output_root', output_root])
  log_file.write(_cur_time()+ 'feature assign start\n')

  try:
    os.system( ' '.join(cmd_list) )
    logger.success('feature_assign--Finished. Find the results in ' + output_root)
  except:
    logger.error('feature_assign--Failed')
  log_file.write(_cur_time()+ 'feature assign finished\n')
  log_file.close()
  _render_template(output_root=output_root, type='quantification', is_server='False')

def pausing_sites(forward_bw=None, reverse_bw=None, output_root=None, cores=1):
  logger.info('pausing_sites--check installed software')
  _check_soft( ['pandas', 'pyBigWig', 'numpy'], isPython=True)

  logger.info('pausing_sites--check parameter and input files')
  if not output_root.endswith('/'): output_root = output_root + '/'
  logger.info('pausing_sites--construct output dir')
  _build_project_path(output_root)
  log_file = open(output_root + 'tmp.log', 'w' )

  cmd_list = ['python', self_scripts + 'pausing_sites_low_memory.py']

  _check_para('--forward_bw', forward_bw, str, isFile=True, required=True, output_root=output_root)
  cmd_list.extend(['--forward_bw', forward_bw])
  log_file.write('forward_bw--'+os.path.basename(forward_bw) )
  log_file.write('forward_bw_size--' + str(os.stat(forward_bw).st_size / (1024 * 1024)) + 'Mb')

  _check_para('--reverse_bw', reverse_bw, str, isFile=True, required=True, output_root=output_root)
  cmd_list.extend(['--reverse_bw', reverse_bw])
  log_file.write('reverse_bw--'+os.path.basename(reverse_bw) )
  log_file.write('reverse_bw_size--' + str(os.stat(reverse_bw).st_size / (1024 * 1024)) + 'Mb')

  cmd_list.extend(['--output_root', output_root])
  if cores:
    cur_cores = _get_cores(cores)
    cmd_list.extend(['--cores', str(cur_cores)])

  log_file.close()
  try:
    os.system( ' '.join(cmd_list) )
    logger.success('pausing_sites--Finished. Find the results in ' + output_root)
  except:
    logger.error('pausing_sites--Failed')
  _render_template(output_root=output_root, type='pausing', is_server='False')

def network_analysis(tf_source=None, tf_filter_nodes=None, enhancer_source=None, enhancer_filter_nodes=None, output_root=None ):
  logger.info('network_analysis--check installed software')
  _check_soft(['networkx','community', 'hvplot', 'numpy'], isPython=True)

  cmd_list = ['python', self_scripts + 'network_analysis.py']

  logger.info('network_analysis--check parameter and input files')
  if not output_root.endswith('/'): output_root = output_root + '/'
  logger.info('network_analysis--construct output dir')
  cmd_list.extend(['--output_root', output_root])
  _build_project_path(output_root)
  log_file = open(output_root + 'tmp.log', 'w' )

  if not(tf_source or enhancer_source):
    logger.error( 'network_analysis--Failed ' + 'no tf or enhancer source file')
    os.sys.exit(1)

  if tf_source:
    _check_para('--tf_source', tf_source, str, isFile=True, required=True, output_root=output_root)
    cmd_list.extend(['--tf_source', tf_source])
  if tf_filter_nodes:
    _check_para('--tf_filter_nodes', tf_filter_nodes, str, isFile=True, required=True, output_root=output_root)
    cmd_list.extend(['--tf_filter_nodes', tf_filter_nodes])
  if enhancer_source:
    _check_para('--enhancer_source', enhancer_source, str, isFile=True, required=True, output_root=output_root)
    cmd_list.extend(['--enhancer_source', enhancer_source])
  if enhancer_filter_nodes:
    _check_para('--enhancer_filter_nodes', enhancer_filter_nodes, str, isFile=True, required=True, output_root=output_root)
    cmd_list.extend(['--enhancer_filter_nodes', enhancer_filter_nodes])

  os.system( ' '.join(cmd_list) )
  logger.success('network analysis--Finished. Find the results in ' + output_root)
  _render_template(output_root=output_root, type='network', is_server='False')

def genome_tracks_visualize(forward_bw=None, reverse_bw=None, region=None, gene=None, output_root=None, cores=1):
  logger.info('genome_tracks_visualize--check installed software')
  _check_soft('pyGenomeTrack', isPython=False)
  logger.info('genome_tracks_visualize--check parameter and input files')
  if not output_root.endswith('/'):
    output_root = output_root + '/'
  logger.info('genome_tracks_visualize--construct output dir')
  _build_project_path(output_root)
  log_file = open(output_root + 'tmp.log', 'w')
  cmd_list = ['bash', self_scripts + 'preprocess.bash']
  cmd_list.extend(['--output_root', output_root])
  # 必须参数 参数什么都不加
  _check_para('--forward_bw', forward_bw, str, isFile=True, required=True, output_root=output_root)
  _check_para('--reverse_bw', reverse_bw, str, isFile=True, required=True, output_root=output_root)

  # check region和gene
  # 生成 pyGenomeTrack ini文件


def _render_template(type='server', output_root=None, is_server=None):
  # 每个步骤对应了输出文件， 用来丰富报表的数据
  # 考虑到 要和论文的内容对应上，所以不能按照 流程来生成报告
  # 把报告分成 1 基础信息 2 qc 3 转录级别评价 4 暂停因子和暂停位点 5 网络分析
  # 只有两种情况 会用到render_template
  logger.info('render_template--check installed software')
  cmd_list = ['python', self_scripts + 'render_template.py']

  cmd_list.extend(['--type', type])
  if not output_root.endswith('/'): output_root = output_root + '/'
  _build_project_path(output_root)
  cmd_list.extend(['--output_root', output_root])
  if is_server:
    cmd_list.extend( ['--is_server', is_server])
  os.system( ' '.join(cmd_list) )


  # 基础信息 table
  # 输入文件名
  # Forward_bw, Reverse_bw,

def main():
  fire.core.Display = lambda lines, out: print(*lines, file=out)
  fire.Fire({
    #'preprocess': preprocess,
    #'alignment': alignment,
    #'genome_tracks': genome_tracks,
    'assessment': assessment,
    'feature_assign': feature_assign,
    'pausing_sites': pausing_sites,
    'network_analysis': network_analysis,
    'render_template': _render_template,
    'server': server,
    'all': all
  })

if __name__ == '__main__':
  main()


import os, fire, re
import pandas as pd
import pyBigWig

import sys
from scipy import stats

# sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))+ '/libs')
script_dir = os.path.abspath(os.path.join(os.path.abspath(__file__), ".."))
lib_dir = os.path.abspath(os.path.join(os.path.abspath(__file__), '../../libs') )
sys.path.append(lib_dir)

## 这个模块可以单独使用
# nasap feature_assign --gtf=./a.gtf --forward=./a.bw --reverse=./b.bw --outpt=./tmp_out

from py_ext import json2dic, dic2json, read_tmp_variable
from parse_gtf import get_gene_df, get_gene_merge_exon_dic, gene_df2dic
from plot import boxplot, scatterplot

def get_total_bases(forward_bw, reverse_bw):
  return pyBigWig.open(forward_bw).header()['sumData'] + pyBigWig.open(reverse_bw).header()['sumData']

def density2rpkm(density, total_bases):
  return density / total_bases * 1e6

def pausing_index( chr, start, end, strand,  bw ):
  # 获取 表达的蛋白， lincRNA
  if strand == '+':
    pp = bw.stats( chr, start, start + 1000)[0]
    gb = bw.stats( chr, start +1000, end)[0]
  else:
    pp = bw.stats( chr, end -1000, end)[0]
    gb = bw.stats( chr, start, end -1000)[0]
  pp_count, gb_count = int(pp* 1000), int(gb* (end-start +1000) )
  if gb == 0:
    #pi = float('nan')
    pi= 'gene body count zero'
  else:
    pi = pp/gb
  return pi, pp_count, gb_count

def elongation_index( chr, start, end, strand,  bw ):
  # 获取 表达的蛋白， lincRNA
  if strand == '+':
    pp = bw.stats( chr, start - 100, start + 300)[0]
    gb = bw.stats( chr, start + 300, start + 2000)[0]
  else:
    pp = bw.stats( chr, end - 300, end + 100)[0]
    gb = bw.stats( chr, end - 2000, end - 300)[0]
  if pp == 0:
    #
    prr = 'proximal promoter count zero'
  else:
    prr = gb/pp
  return prr

def get_attrs(gene_range_dic, forward_bw, reverse_bw, total_bases):
  gene_rpkm_dic, gene_baseCount_dic, gene_pi_dic, gene_ei_dic, gene_pp_count_dic, gene_gb_count_dic = {}, {}, {}, {},{}, {}
  for strand in ['+', '-']:
    strand_gene_range_dic = gene_range_dic[strand]
    bw = pyBigWig.open( {'+': forward_bw, '-': reverse_bw}[strand])
    for gene_name, gene_range in strand_gene_range_dic.items():
      c, s, e = gene_range[0], gene_range[1], gene_range[2]
      try:
        density = bw.stats(c, s, e)[0]
        rpkm = density2rpkm(density, total_bases)
        base_count = int( density * (e - s) )
        if e - s < 2300:
          pi, prr='gene too short', 'gene too short'
          continue
        else:
          pi, pp_count, gb_count = pausing_index(c,s, e, strand, bw)
          prr = elongation_index(c,s, e, strand, bw)

      except:
        continue
      gene_rpkm_dic[gene_name] = rpkm
      gene_baseCount_dic[gene_name] = base_count
      gene_pi_dic[gene_name] = pi
      gene_ei_dic[gene_name] = prr
      gene_pp_count_dic[gene_name] = pp_count
      gene_gb_count_dic[gene_name] = gb_count
    # dic2json(gene_rpkm_dic, rpkm_json_output)
  return gene_rpkm_dic, gene_baseCount_dic, gene_pi_dic, gene_ei_dic, gene_pp_count_dic, gene_gb_count_dic

# count, RPKM, PI, EI, Exon/intron density
def main(gtf, forward_bw, reverse_bw, output_root ):
  total_bases = get_total_bases(forward_bw, reverse_bw)

  gene_df = get_gene_df( gtf )
  filter_gene_df = gene_df[ gene_df['genetype'].isin(['protein_coding', 'lincRNA', 'lncRNA']) ]
  gene_range_dic = gene_df2dic(filter_gene_df)
  # count, RPKM, promoter, geneBody count, PI, EI

  gene_rpkm_dic, gene_baseCount_dic, gene_pi_dic, gene_ei_dic, gene_pp_count_dic, gene_gb_count_dic = get_attrs(gene_range_dic, forward_bw, reverse_bw, total_bases)
  # print( list(gene_rpkm_dic.items())[:5] )

  attr_list = ['baseCount', 'rpkm', 'pp_count', 'gb_count', 'pi', 'ei']
  for gene_type in set(filter_gene_df['genetype']):
    gene_list = filter_gene_df[filter_gene_df['genetype'] == gene_type]['gene_name']
    for attr in attr_list:
      # type_attr_dic = {g: globals()['gene_' + attr +'_dic'][g] for g in gene_list}
      tmp_dic = eval('gene_' + attr +'_dic')
      tmp_list = list( tmp_dic.keys() )
      type_attr_dic = {g: tmp_dic[g] for g in gene_list if g in tmp_list}
      pd.DataFrame( {attr: pd.Series(type_attr_dic)} ).to_csv(output_root+'csv/' + gene_type + '_'+ attr + '.csv')


  gene_attrs_df = pd.concat({'rpkm':pd.Series(gene_rpkm_dic), 'count': pd.Series(gene_baseCount_dic),
   'pp_count': pd.Series(gene_pp_count_dic), 'gb_count': pd.Series(gene_gb_count_dic),
   'pi': pd.Series(gene_pi_dic), 'ei': pd.Series(gene_ei_dic)}, axis=1)

  gene_attrs_df.to_csv(output_root+'csv/all_feature_attrs.csv')

  chr_list = list( set( filter_gene_df['chrom'] ) )

  sort_chr_list = [chrom for chrom in chr_list if re.match( r'chr\d+', chrom)]
  sort_chr_list.sort(key=lambda arr: (arr[:3], int(arr[3:])))
  if 'chrM' in chr_list:  sort_chr_list.append('chrM')
  if 'chrMT' in chr_list:  sort_chr_list.append('chrMT')
  if 'chrX' in chr_list:  sort_chr_list.append('chrX')
  if 'chrY' in chr_list:  sort_chr_list.append('chrY')
  rpkm_list = []
  for chrom in sort_chr_list:
    gene_list = filter_gene_df[ filter_gene_df['chrom'] == chrom ]['gene_name']

    cur_rpkm_list = []
    for gene in gene_list:
      try:
        cur_rpkm_list.append(gene_rpkm_dic[gene])
      except:
        continue
    rpkm_list.append( cur_rpkm_list )
  boxplot(rpkm_list, 'chromsome genes RPKM', 'chromsome', 'RPKM', sort_chr_list, output_root + 'imgs/chr_rpkm.png' )



  # Exon/Intron
  protein_exon_range_dic = get_gene_merge_exon_dic(gtf)
  forward_bw_signal = pyBigWig.open(  forward_bw )
  reverse_bw_signal = pyBigWig.open(  reverse_bw )
  exon_intron_ratio_dic = {}
  for gene_name, gr_dic in protein_exon_range_dic.items():
    chrom, strand, exon_range_list = gr_dic['chrom'], gr_dic['strand'], gr_dic['exon']
    try:
      gene_range = gene_range_dic[strand][gene_name]
    except:
      print('no gene', gene_name)
      os.sys.exit(0)
      continue
    gene_start, gene_end = gene_range[1], gene_range[2]
    if strand == '+':
      bw = forward_bw_signal
    else:
      bw = reverse_bw_signal
    gene_length = gene_end - gene_start
    try:
      gene_total = bw.stats(chrom, gene_start, gene_end)[0] * gene_length
    except:
      # print(gene_name, chrom, gene_start, gene_end)
      continue
    exon_length, exon_total = 0, 0
    for exon_range in exon_range_list:
      cur_length = (exon_range[1] - exon_range[0])
      exon_length+=cur_length
      try:
        exon_total +=( cur_length* bw.stats(chrom, exon_range[0], exon_range[1])[0] )
      except:
        continue
    intron_total = gene_total - exon_total
    intron_length = gene_length - exon_length

    if intron_total <= 0:
      # print( "warning: intron density <=0")
      exon_density, intron_density,  ratio = exon_total/exon_length, 0, 0
    else:
      exon_density, intron_density = exon_total/exon_length, intron_total/intron_length
      ratio = exon_density/intron_density
    exon_intron_ratio_dic[gene_name] = {'exon': exon_density, 'intron': intron_density, 'ratio': ratio}

  exon_intron_df = pd.DataFrame(exon_intron_ratio_dic)
  exon_intron_df = exon_intron_df.T
  exon_intron_df.to_csv(output_root+'csv/exon_intron_ratio.csv')

  filter_exon_intron_df = exon_intron_df[(exon_intron_df['exon'] > 0) & (exon_intron_df['intron'] >0) ]


  slope, intercept, r_value, p_value, std_err = stats.linregress(filter_exon_intron_df['exon'],filter_exon_intron_df['intron'])
  text= 'y = {:4.2e} x + {:4.2e}; \nR^2= {:2.2f}'.format(slope, intercept, r_value*r_value)
  scatterplot( filter_exon_intron_df['exon'], filter_exon_intron_df['intron'], 'Exon/Intron ratio', 'Exon', 'Intron', text, output_root+'imgs/exon_intron_ratio.png' )

if __name__ == '__main__':
  fire.Fire( main )


import os, sys
from collections import Counter

import matplotlib.pyplot as plt
from matplotlib.pyplot import Polygon

script_dir = os.path.abspath(os.path.join(os.path.abspath(__file__), ".."))
lib_dir = os.path.abspath(os.path.join(os.path.abspath(__file__), '../../libs') )
sys.path.append(lib_dir)

from py_ext import json2dic, get_file_size

output_root = sys.argv[1]
if not output_root.endswith('/'): output_root = output_root +'/'
# 目标:
# 1 生成 preprocess_report.txt
# 2 生成 3张图片

global stats_list
stats_list = []

def parse_fastp_json(json_dir):
  fastp_dic = json2dic(json_dir)
  summary = fastp_dic['summary']
  before = summary['before_filtering']
  after = summary['after_filtering']
  res = fastp_dic['filtering_result']

  para_dic = {
   'before_reads': int(before['total_reads']),
   'before_bases': int(before['total_bases']),
   'after_reads': int(after['total_reads']),
   'after_bases': int(after['total_bases']),

   'before_q20_rate': float(before['q20_rate']),
   'before_q30_rate': float(before['q30_rate']),
   'before_read1_mean_length': float(before['read1_mean_length']),
   'before_gc_rate': float(before['gc_content']),
   'after_q20_rate': float(after['q20_rate']),
   'after_q30_rate': float(after['q30_rate']),
   'after_reads_mean_length': float(after['read1_mean_length']),
   'after_gc_rate': float(after['gc_content']),

   'passed_filter_reads': int(res['passed_filter_reads']),
   'low_quality_reads': int(res['low_quality_reads']),
   'too_short_reads': int(res['too_short_reads']),
   'too_many_N_reads': int(res['too_many_N_reads'])
  }
  if 'read2_mean_length' in before.keys():
    para_dic['before_read2_mean_length'] = float(before['read2_mean_length'])
  if 'read2_mean_length' in after.keys():
    para_dic['after_read2_mean_length'] = float(after['read2_mean_length'])
  if 'adapter_cutting' in fastp_dic.keys():
    para_dic['adapter_trimmed_reads'] = fastp_dic['adapter_cutting']['adapter_trimmed_reads']
  if 'polyx_trimming' in fastp_dic.keys():
    para_dic['polyx_trimmed_reads'] = fastp_dic['polyx_trimming']['total_polyx_trimmed_reads']
  return para_dic


# 0 get variable
tmp_variable_dir = output_root+"tmp_variable.txt"
variable_dic = {ln.split('--')[0]: ln.split('--')[1].strip() for ln in open(tmp_variable_dir)}
stats_list.append( ['Read1_name', os.path.basename(variable_dic['read1_name'])] )
stats_list.append( ['Read1_num', variable_dic['read1_num']] )
stats_list.append( ['Read1_size',get_file_size(variable_dic['read1_name'])] )
try:
  stats_list.append( ['Read2_name', os.path.basename(variable_dic['read2_name'])] )
  stats_list.append( ['Read2_num', variable_dic['read2_num']] )
  stats_list.append( ['Read2_size',get_file_size(variable_dic['read2_name'])] )
except:
  pass

stats_list.append( ['Reads_with_adapter', variable_dic['reads_with_adapter']] )
stats_list.append( ['Uninformative_adapter_reads', variable_dic['uninformative_adapter_reads']] )
stats_list.append( ['Pct_uninformative_adapter_reads', variable_dic['pct_uninformative_adapter_reads']] )
stats_list.append( ['Peak_adapter_insertion_size', variable_dic['peak_adapter_insertion_size'] ] )

remove_adapter_dic = parse_fastp_json(output_root+"json/remove_adapter.json")
cut_twoEnd_dic = parse_fastp_json(output_root+"json/trim.json")
filter_quality_dic = parse_fastp_json(output_root+"json/filter_quality.json")
remove_polyX_dic = parse_fastp_json(output_root+"json/remove_polyX.json")

# 1.1 raw reads
total_reads = remove_adapter_dic['before_reads']
s1_1 = total_reads
s1_2, s1_3 = 0, 0
annotate_text1 = 'Total reads: {reads}\nTotal bases: {bases}\nReads mean length: {mean_length}\nQ20 rate: {q20}\nQ30 rate: {q30}\nGC rate: {gc}'.format(
  reads=total_reads, bases=remove_adapter_dic['before_bases'], mean_length=remove_adapter_dic['before_read1_mean_length'],
  q20=remove_adapter_dic['before_q20_rate'], q30=remove_adapter_dic['before_q30_rate'], gc=remove_adapter_dic['before_gc_rate']
)

# 1.2 remove adapter
pass_adapter = remove_adapter_dic['passed_filter_reads']
fail_adapter = remove_adapter_dic['too_short_reads']
with_adapter = remove_adapter_dic['adapter_trimmed_reads']
s2_1 = with_adapter
s2_2 = pass_adapter - with_adapter
s2_3 = fail_adapter
annotate_text2 = 'Total reads: {reads}\nTotal bases: {bases}\nReads mean length: {mean_length}\nQ20 rate: {q20}\nQ30 rate: {q30}\nGC rate: {gc}'.format(
  reads=pass_adapter, bases=remove_adapter_dic['after_bases'], mean_length=remove_adapter_dic['after_reads_mean_length'],
  q20=remove_adapter_dic['after_q20_rate'], q30=remove_adapter_dic['after_q30_rate'], gc=remove_adapter_dic['after_gc_rate']
)
stats_list.append( ['Reads_with_adapter', with_adapter] )
stats_list.append( ['Uninformative_adapter_reads', fail_adapter] )
stats_list.append( ['Pct_uninformative_adapter_reads', round(fail_adapter/(pass_adapter+fail_adapter)*100, 3)] )
# stats_list.append( ['Peak_adapter_insertion_size', peak_adapter_insertion_size] )
before_adapter_bases = remove_adapter_dic['before_bases']
after_adapter_bases = remove_adapter_dic['after_bases']
adapter_loss_rate = (before_adapter_bases - after_adapter_bases)/before_adapter_bases
stats_list.append( ['Adapter_loss_rate', adapter_loss_rate] )

def draw_adapter_distribution(output_root):
  # x->insertion size,
  # y->number of reads,
  # text->degradation ratio, x-label, y-label
  # scatter
  # 两段阴影
  insertion_size_list = [int( ln.split('\t')[1].strip() ) for ln in open(output_root + 'txt/adapter_read_len1.txt')]
  # 计算 y->insertion size的频率(Counter就行)
  insert_size_counter = Counter(insertion_size_list)
  x_list = list( insert_size_counter.keys() )
  y_list = list( insert_size_counter.values() )
  x_max = max(x_list)  # x_max对应reads 没有去掉 adapter
  adapter_length, length_count = [], []
  degraded_reads, intact_reads = 0, 0
  for x, y in zip(x_list, y_list):
    if x==x_max:
      continue
    l = x_max - x
    adapter_length.append( l )
    length_count.append(y)
    if x >= 10 and x < 20:
      degraded_reads+=y
    if x >= 30 and x < 40:
      intact_reads+=y

  degradation_ratio = degraded_reads/float(intact_reads)
  stats_list.append( ['Degradation ratio', degradation_ratio] )

  fig,ax = plt.subplots()
  max_length =max(length_count)
  ax.set_ylim(ymin=0, ymax= max_length )

  # 散点
  # plt.scatter(x_list,y_list )
  ax.scatter(adapter_length,length_count, color='black', alpha=0.6 )
  # 阴影 用polygon
  poly1=Polygon([(0, 0),(0, max_length*1.1), (20, max_length*1.1), (20, 0)],alpha=0.3, facecolor='#FFE6D0', edgecolor="black", linestyle="-." )
  poly2=Polygon([(20, 0), (20, max_length*1.1), (30, max_length*1.1), (30, 0)],alpha=0.3, facecolor='#FFFDE5', edgecolor="black", linestyle="-."  )
  ax.add_patch(poly1)
  ax.add_patch(poly2)

  # text
  # plt.text(0.5, 0.5, 'degradation rate' + str(round(degradation_ratio, 2) ) )
  plt.text(10, int(max_length/3),  'high degradation', fontsize=12, rotation='vertical', zorder=99)
  plt.text(25, int(max_length/3),  'partial degradation', fontsize=12, rotation='vertical', zorder=99)
  plt.text(0.7, 0.9, 'degradation ratio ' + str(round(degradation_ratio, 2) ) , ha='center', va='center', transform=ax.transAxes, zorder=99, backgroundcolor='white')
  plt.xlim([0, max(adapter_length)])
  plt.ylim([0, max_length * 1.1])
  plt.xlabel('Size of insertion')
  plt.ylabel('Number of reads')
  plt.savefig(output_root + 'imgs/adapter_insertion_distribution.png' )
  plt.savefig(output_root + 'imgs/adapter_insertion_distribution.pdf' )

# adapter_dist_list = defaultdict(lambda: [])
# for ln in open(output_root + '/adapter_distribution.txt'):
#   ls = ln.split('\t')
#   adapter_dist_list[ls[0]].append( int(ls[1]) )

# adapter_dist_dic = {k: abs(v[0]-v[1]) for k, v in adapter_dist_list.items()  if len(v) == 2}
# # 这里的peak，我认为是 数量最多的 insertion 长度
# peak_adapter_insertion_size = max( adapter_dist_dic.values() )
# print('peak_adapter_insertion_size', peak_adapter_insertion_size )
# adapter_length_list = list(adapter_dist_dic.values())
draw_adapter_distribution( output_root  )

# 1.3 trim two ends
pass_twoEnd = cut_twoEnd_dic['passed_filter_reads']
fail_twoEnd = cut_twoEnd_dic['too_short_reads']
with_twoEnd = int(variable_dic['reads_with_cutTwoEnd'])
s3_1 = with_twoEnd
s3_2 = pass_twoEnd - with_twoEnd
s3_3 = fail_twoEnd
annotate_text3 = 'Total reads: {reads}\nTotal bases: {bases}\nReads mean length: {mean_length}\nQ20 rate: {q20}\nQ30 rate: {q30}\nGC rate: {gc}'.format(
  reads=pass_twoEnd, bases=cut_twoEnd_dic['after_bases'], mean_length=cut_twoEnd_dic['after_reads_mean_length'],
  q20=cut_twoEnd_dic['after_q20_rate'], q30=cut_twoEnd_dic['after_q30_rate'], gc=cut_twoEnd_dic['after_gc_rate']
)

stats_list.append( ['Trimmed_reads', with_twoEnd ])
before_trim_bases = cut_twoEnd_dic['before_bases']
after_trim_bases = cut_twoEnd_dic['after_bases']
trim_loss_rate= round( (before_trim_bases-after_trim_bases)/ before_trim_bases*100, 2)
stats_list.append( ['Trim_loss_rate', trim_loss_rate] )

# 1.4 remove polyX
pass_polyX = remove_polyX_dic['passed_filter_reads']
fail_polyX = remove_polyX_dic['too_short_reads']
with_polyX = remove_polyX_dic['polyx_trimmed_reads']
s4_1 = with_polyX
s4_2 = pass_polyX - with_polyX
s4_3 = fail_polyX
annotate_text4 = 'Total reads: {reads}\nTotal bases: {bases}\nReads mean length: {mean_length}\nQ20 rate: {q20}\nQ30 rate: {q30}\nGC rate: {gc}'.format(
  reads=pass_polyX, bases=remove_polyX_dic['after_bases'], mean_length=remove_polyX_dic['after_reads_mean_length'],
  q20=remove_polyX_dic['after_q20_rate'], q30=remove_polyX_dic['after_q30_rate'], gc=remove_polyX_dic['after_gc_rate']
)
stats_list.append( ['Reads_with_polyX', with_polyX] )
stats_list.append( ['Uninformative_polyX_reads', fail_polyX] )

# 1.5 mean quality
pass_quality = filter_quality_dic['passed_filter_reads']
fail_quality = filter_quality_dic['low_quality_reads'] + filter_quality_dic['too_many_N_reads']
s5_1 = pass_quality
s5_2 = 0
s5_3 = fail_quality
annotate_text5 = 'Total reads: {reads}\nTotal bases: {bases}\nReads mean length: {mean_length}\nQ20 rate: {q20}\nQ30 rate: {q30}\nGC rate: {gc}'.format(
  reads=pass_quality, bases=filter_quality_dic['after_bases'], mean_length=filter_quality_dic['after_reads_mean_length'],
  q20=filter_quality_dic['after_q20_rate'], q30=filter_quality_dic['after_q30_rate'], gc=filter_quality_dic['after_gc_rate']
)

stackbar_list = [
  # have not failed
  [s1_1, s2_1, s3_1, s4_1, s5_1],
  [s1_2, s2_2, s3_2, s4_2, s5_2],
  [s1_3, s2_3, s3_3, s4_3, s5_3]
]
def draw_readsNum(steps, stackbar_list):
  # 多堆叠图 [total-> keep(with or not), failed]
  # 同样的y ['original', 'quantity filter', 'adapter', 'trim two side', 'polyX']
  # 堆叠图 有多个x, x_have, x_not, x_failed
  x_have, x_not, x_failed = stackbar_list
  plt.figure(figsize=(10,7))#设置画布的尺寸
  plt.bar(steps, x_have, label="reads with feature",edgecolor = 'black')
  plt.bar(steps, x_not, label="reads without feature",edgecolor = 'black', bottom = x_have)
  plt.bar(steps, x_failed, label="failed reads",edgecolor = 'black', bottom = [i+j for i, j in zip(x_have,x_not)])
  plt.legend( loc=3,fontsize=14)  # 设置图例位置
  plt.ylabel('Reads number',fontsize=16)
  plt.xlabel('Preprocess steps',fontsize=16)
  # plt.title("stack plot",fontsize=18 )

  for x1, y1, y2, y3 in zip(steps, x_have, x_not, x_failed):
    p1 = y1/(y1+y2+y3)
    p2 = y2/(y1+y2+y3)
    p3 = y3/(y1+y2+y3)
    if p1 >0.05:
      plt.text(x1, y1 * 0.4, '{:.0%}'.format(p1), ha='center',fontsize = 15)
    if p2>0.05:
      plt.text(x1, y1 + (y2)* 0.4,  '{:.0%}'.format(p2), ha='center',fontsize = 15)
    if p3>0.05:
      plt.text(x1, y1 + y2 + (y3)* 0.4, '{:.0%}'.format(p3), ha='center',fontsize = 15)

  # plt.show()
  plt.savefig(output_root +'imgs/reads_ratio.png')
  plt.savefig(output_root +'imgs/reads_ratio.pdf')

steps=['raw reads', 'remove adapter', 'trim two ends', 'remove polyX', 'filter quality']
draw_readsNum(steps, stackbar_list)


def reads_length_dist(reads_len_file, min, max):
  reads_length = [int(ln.split('\t')[1]) for ln in open(reads_len_file) ]
  length_counter = Counter(reads_length)
  x = list( range(min, max+1) )
  y = [length_counter[i] if i in length_counter.keys() else 0 for i in x ]
  return [x, y]

read_max_length = int( remove_adapter_dic['before_read1_mean_length'] )

remove_adapter_list = reads_length_dist(output_root + "/txt/adapter_read_len1.txt", 1, read_max_length)
cut_twoEnd_list = reads_length_dist(output_root + "/txt/filter_trim_len1.txt", 1, read_max_length)
remove_polyX_list = reads_length_dist(output_root + "/txt/filter_polyX_len1.txt", 1, read_max_length)
filter_quality_list = reads_length_dist(output_root + "/txt/filter_quality_len1.txt", 1, read_max_length)

bar_list = [[list(range(1, read_max_length+1)), [0]*(read_max_length-1) + [total_reads]], filter_quality_list, remove_adapter_list, remove_polyX_list, cut_twoEnd_list]

def readsLength_dist_subplots(steps, bar_list, text_list): #
  fig, axes = plt.subplots(len(steps), 1, sharex=True, sharey=True, figsize=(14, 20))
  for i, step in enumerate(steps):
    x, y = bar_list[i]
    axes[i].bar(x, y)
    axes[i].set_ylabel('log10(Reads number)' )
    axes[i].set_title('Preprocess: ' + step, fontsize = 14, fontweight ='bold', verticalalignment='bottom', loc ='left')
    axes[i].text(0.01, 0.9, text_list[i],
        horizontalalignment='left',
        verticalalignment='top',
        transform=axes[i].transAxes)
  plt.xlabel('Reads length')
  plt.yscale("log")
  # plt.show()
  plt.savefig(output_root + 'imgs/reads_distribution.png')
  plt.savefig(output_root + 'imgs/reads_distribution.pdf')

readsLength_dist_subplots(steps, bar_list, [annotate_text1, annotate_text2, annotate_text3, annotate_text4, annotate_text5])


c = open(output_root + 'csv/preprocess_report.csv', 'w')
for stat_item in stats_list:
  c.write(stat_item[0] + ',' + str(stat_item[1]) + '\n')
c.close()





import string

def render( para_dic ): # , output_root):
  template_str = open('./template.html').read()

  for k, v in para_dic.items():
    print( str( k ) )
    template_str = template_str.replace( str('$'+k),  str(v) )
  print( template_str)

render( {'who': 'biodancer'})



import sys, os
from glob import glob
from setuptools import setup,find_packages

def main():
  root = os.path.abspath(os.path.dirname(__file__))
  with open(os.path.join(root, 'requirements.txt')) as f:
    install_requires = f.read().splitlines()
  # linux depandance packages
  # linux_deps = ['fastp', 'bioawk', 'python', 'bowtie2', 'samtools', 'bedtools', 'deeptools']
  # for dep in linux_deps:
  #   is_dep = os.system(dep + ' --version')
  #   if is_dep != 0:
  #     apt install

  if float(sys.version[:3])<=3.6:
    sys.stderr.write("CRITICAL: Python version must be >= 3.6x!\n")
    sys.exit(1)

  setup(
    name='nasap',
    version='0.2.0',
    description='This is nASAP setup file',
    author='biodancer',
    author_email='szxszx@foxmail.com',
    url='https://github.com/biodancerwanghzi/nasap/',
    # packages= find_packages(exclude=["back"]),
    packages=['src'],
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    data_files= [('src/scripts/', glob('src/scripts/[!_]*.py', recursive=True) ),
    ('src/scripts/templates/', ['src/scripts/templates/template.html', 'src/scripts/templates/template_track.txt']),
    ('src/scripts/templates/static/', ['src/scripts/templates/static/vue.min.js'])
    ],
    entry_points={
      'console_scripts': [
        'nasap=src.nasap:main',
        'batch_nasap=src.batch_nasap:main'
      ]
    },
    install_requires=install_requires,
    extras_require={
      'genomeBrowser': [
        'pyGenomeTrack'
    ]

    },
    zip_safe = True
  )

if __name__ == '__main__':
  main()